import sys
from typing import Any

class _PyrionBaseType(int):
    __slots__ = ()

    def __new__(cls, value, mask=0xFFFFFFFF):
        # Mask
        value = int(value) & mask
        return super().__new__(cls, value)

    def __add__(self, other):
        return _PyrionBaseType(int(self) + int(other))

    def __sub__(self, other):
        return _PyrionBaseType(int(self) - int(other))

    def __mul__(self, other):
        return _PyrionBaseType(int(self) * int(other))

    def __and__(self, other):
        return _PyrionBaseType(int(self) & int(other))

    def __or__(self, other):
        return _PyrionBaseType(int(self) | int(other))

    def __xor__(self, other):
        return _PyrionBaseType(int(self) ^ int(other))

    def __invert__(self):
        return _PyrionBaseType(~int(self))

class pyrion_uint64(_PyrionBaseType):
    """Represents unsigned integer of size 64-bits"""
    def __new__(cls, value):
        return super().__new__(cls, value, 0xFFFFFFFFFFFFFFFF)

    def __repr__(self):
        return f"pyrion_uint64({int(self)})"

class pyrion_uint32(_PyrionBaseType):
    """Represents unsigned integer of size 32-bits"""
    def __new__(cls, value):
        return super().__new__(cls, value, 0xFFFFFFFF)

    def __repr__(self):
        return f"pyrion_uint32({int(self)})"
    
class pyrion_uint16(_PyrionBaseType):
    """Represents unsigned integer of size 16-bits"""
    def __new__(cls, value):
        return super().__new__(cls, value, 0xFFFF)

    def __repr__(self):
        return f"pyrion_uint16({int(self)})"
    
class pyrion_uint8(_PyrionBaseType):
    """Represents unsigned integer of size 8-bits"""
    def __new__(cls, value):
        return super().__new__(cls, value, 0xFF)

    def __repr__(self):
        return f"pyrion_uint8({int(self)})"

class pyrion_int64(_PyrionBaseType):
    """Represents signed integer of size 64-bits"""
    def __new__(cls, value):
        value = int(value) & 0xFFFFFFFFFFFFFFFF
        # convert to signed range
        if value & 0x8000000000000000:
            value -= 0x10000000000000000
        return super().__new__(cls, value, 0xFFFFFFFFFFFFFFFF)

    def __repr__(self):
        return f"pyrion_int64({int(self)})"

class pyrion_int32(_PyrionBaseType):
    """Represents signed integer of size 32-bits"""
    def __new__(cls, value):
        value = int(value) & 0xFFFFFFFF
        if value & 0x80000000:
            value -= 0x100000000
        return super().__new__(cls, value, 0xFFFFFFFF)

    def __repr__(self):
        return f"pyrion_int32({int(self)})"
    
class pyrion_int16(_PyrionBaseType):
    """Represents signed integer of size 16-bits"""
    def __new__(cls, value):
        value = int(value) & 0xFFFF
        if value & 0x8000:
            value -= 0x10000
        return super().__new__(cls, value, 0xFFFF)

    def __repr__(self):
        return f"pyrion_int16({int(self)})"
    
class pyrion_int8(_PyrionBaseType):
    """Represents signed integer of size 8-bits"""
    def __new__(cls, value):
        value = int(value) & 0xFF
        if value & 0x80:
            value -= 0x100
        return super().__new__(cls, value, 0xFF)

    def __repr__(self):
        return f"pyrion_int8({int(self)})"

if sys.maxsize == 2**32:
    _BITS = 64
else:
    _BITS = 32

class pyrion_void:
    """Represents nothing, just an empty void"""
    __slots__ = ()

    def __repr__(self):
        return "pyrion_void()"
    
class _PyrionPointer:
    """Base class for Pyrion pointers."""
    __slots__ = ('address',)

    def __init__(self, address: int = 0):
        self.address = pyrion_uint64(address) if _BITS == 64 else pyrion_uint32(address)
        
    def __add__(self, other):
        return _PyrionPointer(int(self.address) + int(other))

    def __sub__(self, other):
        return _PyrionPointer(int(self.address) - int(other))

    def __repr__(self):
        return f"{self.__class__.__name__}({hex(self.address)})"

    def is_null(self):
        return self.address == 0
    
    def __eq__(self, other):
        return isinstance(other, _PyrionPointer) and self.__address == other.__address

    @classmethod
    def __class_getitem__(cls, *args):
        return cls

class pyrion_ptr(_PyrionPointer):
    """Generic typed pointer."""
    __slots__ = ('ptype', "_value")

    def __init__(self, ptype, value: Any=None):
        super().__init__(id(value))
        self.ptype = ptype  # what type this pointer points to
        self._value = value

    def cast(self, new_type):
        """Return a new pointer of different type."""
        return pyrion_ptr(new_type, self.address)

    def __repr__(self):
        return f"pyrion_ptr<{self.ptype.__name__}>({hex(self.address)})"
    
    @property
    def value(self) -> Any:
        return self._value
       
    def check_if_same(self, other):
        """Checks if the VALUE stored on both pointers are same"""
        return isinstance(other, pyrion_ptr) and self.value == other.value

class pyrion_structure_meta(type):
    """Meta for pyrion_structure to handle fields."""
    def __new__(cls, name, bases, namespace):
        fields = namespace.get("_struct_", [])
        size = 0
        offsets = {}

        # Compute offsets & total size
        for field in fields:
            if len(field) == 2:
                fname, ftype = field
                fsize = getattr(ftype, "_size_", 4)
            elif len(field) == 3:  # array
                fname, ftype, flen = field
                fsize = getattr(ftype, "_size_", 4) * flen
            else:
                raise ValueError("Invalid field tuple")
            offsets[fname] = size
            size += fsize

        namespace["_size_"] = size
        namespace["_offsets_"] = offsets
        return super().__new__(cls, name, bases, namespace)

class pyrion_structure(metaclass=pyrion_structure_meta): # Works like ctypes.Structure, so credits!
    _struct_ = []

    def __init__(self, mem: bytearray = None, offset: int = 0):
        self._mem = mem if mem is not None else bytearray(self._size_)
        self._offset = offset

    def __getattr__(self, name):
        if name in self._offsets_:
            offset = self._offsets_[name] + self._offset
            ftype = dict(self._struct_)[name]
            fsize = getattr(ftype, "_size_", 4)
            val_bytes = self._mem[offset:offset+fsize]
            # convert bytes to int for basic types
            if issubclass(ftype, _PyrionBaseType):
                return ftype(int.from_bytes(val_bytes, "little"))
            return val_bytes
        raise AttributeError(f"{name} not found")

    def __setattr__(self, name, value):
        if name in ("_mem", "_offset", "_struct_", "_size_", "_offsets_"):
            super().__setattr__(name, value)
        elif name in self._offsets_:
            offset = self._offsets_[name] + self._offset
            ftype = dict(self._struct_)[name]
            fsize = getattr(ftype, "_size_", 4)
            if issubclass(ftype, _PyrionBaseType):
                self._mem[offset:offset+fsize] = int(value).to_bytes(fsize, "little")
            else:
                # raw bytes assignment
                self._mem[offset:offset+fsize] = bytes(value)
        else:
            raise AttributeError(f"{name} not found")



import sys
import threading
import inspect
import queue
import socket

class MemBlock:
    """A Memory Block class used to define protected sections of RAM (Virtual RAM)"""
    def __init__(self, size:int, access:list[str], name:str="BLOCK", seek:int=0):
        """Init func

        Args:
            size (int): The size of the protected region in bytes.
            access (str): The __file__ variable of the file that can access this region.
            name (str, optional): Name of the region. Defaults to "BLOCK".
            seek (int, optional): Address where the region starts. Defaults to 0.
        """
        self.size = size
        self.access = access
        self.name = name
        self.seek = seek
        self.end_seek = seek + size

class Process_Data:
    """A Class for VRam Mapping Data for processes
    """
    def __init__(self, Pid:int, process:object, *params):
        """The init func

        Args:
            Pid (int): The Unique ID given to every func (Process ID).
            process (object): The function to run as a process.
            *params: Params to the process.
        """
        self.Pid = Pid
        self.process:object = process
        self.params = params

class Execution_Data:
    """Execution Data Class
    """
    def __init__(self, heap_size:int=6, stack_size:int=6, heap_reserve:int=1, stack_reserve:int=1, data_size:int=8) -> None:
        """Init func

        Args:
            heap_size (int, optional): MAX Heap size including Heap Reserve in bytes. Defaults to 6.
            stack_size (int, optional): MAX Stack size including Stack Reserve in bytes. Defaults to 6.
            heap_reserve (int, optional): Reserved Heap in bytes. Defaults to 1.
            stack_reserve (int, optional): Reserved Stack in bytes. Defaults to 1.
            data_size (int, optional): Size of initialized data in bytes. Defaults to 8.
        """

        self.heap_reserve:int = heap_reserve
        self.stack_reserve:int = stack_reserve
        self.heap_size:int = heap_size
        self.stack_size:int = stack_size
        self.data_size:int = data_size
        self.vmem_size:int = heap_size + stack_size + data_size

class Memory:
    """A Memory Class.

    Memory Layout ->

    System Memory -> 64 bytes (Blocked)

    Process Memory -> User-Defined, Default 64 bytes (Open)

    -- Program Memory (Virtual Memory) -> Per Process , User - Defined (Blocked)

    Rest of the Memory -> User-Defined (Open)
    """
    def __init__(self, size:int, proc_sector_size:int=64, Block:list[MemBlock]=None, debug:bool=False, system_size:int=64) -> None:
        """Initialization function.
        NOTE: RAM in this class refers to a part of mapped RAM known as Virtual Memory that every program has.

        Args:
            size (int): The size of RAM in bytes.
            proc_sector_size (int): The size of RAM dedicated to processes in bytes. Defaults to 64.
            Block (list[MemBlock]): The protected regions if any. Defaults to None.
        """
        self.size:int = size + system_size + proc_sector_size# Size + System Size + Process Sector Size
        self.ram:bytearray = bytearray(size)
        self.current_addr:int = 0x0
        self.mem_blocks:list[MemBlock] = Block

        self.proc_sector_size = proc_sector_size

        self.sys_block:MemBlock = MemBlock(system_size, [__file__], "_WIN32-PR", 0)

        self.debug = debug

        if self.mem_blocks:
            if self._init_blocks() == False:
                exit()

    def _get_running_file(self) -> None:
        """DO NOT USE!"""
        import inspect
        stack = inspect.stack()

        for frame in stack[1:]:  # Skip self
            caller_locals = frame.frame.f_locals
            if "self" in caller_locals:
                if caller_locals["self"].__class__ == self.__class__:
                    continue  # Skip internal calls within the same class
            return frame.filename  # First external caller

        return None  # No external caller found (edge case)

    def _init_sys_ram(self) -> None:
        """DO NOT USE!"""
        pass

    def _init_blocks(self) -> bool:
        """DO NOT USE!"""
        for block_p in self.mem_blocks:
            addr = block_p.seek
            size = block_p.size

            end_seek = addr + size

            if addr >= self.sys_block.seek and addr <= self.sys_block.end_seek:
                if self.debug: print(f"PHardwareITK/Memory/_platform/_win32: ERROR -> block named [{block_p.name}] overlapes a system block named -> {self.sys_block.name}")
                return False

            for block in self.mem_blocks:
                if block_p == block:
                    continue

                if addr >= block.seek and addr <= block.end_seek:
                    if self.debug: print(f"PHardwareITK/Memory/_platform/_win32: ERROR -> block named [{block_p.name}] overlapes a block named -> {block.name}")
                    return False

                if end_seek >= block.seek and end_seek <= block.end_seek:
                    if self.debug: print(f"PHardwareITK/Memory/_platform/_win32: ERROR -> block named [{block_p.name}] overlapes a block named -> {block.name}")
                    return False

                if addr < block.seek and end_seek > block.end_seek:
                    if self.debug: print(f"PHardwareITK/Memory/_platform/_win32: ERROR -> block named [{block_p.name}] overlapes a block named -> {block.name}")
                    return False

        return True

    def _CFBlock(self, addr:int, size:int) -> bool:
        file =  self._get_running_file()

        end_seek = addr + size

        if addr >= self.sys_block.seek and addr <= self.sys_block.end_seek:
            if not self.sys_block.access == file and not self.sys_block.size == 0:
                if self.debug: print(f"PHardwareITK/Memory/_platform/_win32: ERROR -> Address comes under a system block named -> {self.sys_block.name}")
                return False

        if not self.mem_blocks == None:
            for block in self.mem_blocks:
                if addr >= block.seek and addr <= block.end_seek:
                    if not file in block.access and not block.size == 0:
                        if self.debug: print(f"PHardwareITK/Memory/_platform/_win32: ERROR -> Address comes under a block named -> {block.name}")
                        return False

                if end_seek >= block.seek and end_seek <= block.end_seek:
                    if not file in block.access and not block.size == 0:
                        if self.debug: print(f"PHardwareITK/Memory/_platform/_win32: ERROR -> End of Data comes under a block named -> {block.name}")
                        return False

                if addr < block.seek and end_seek > block.end_seek:
                    if not file in block.access and not block.size == 0:
                        if self.debug: print(f"PHardwareITK/Memory/_platform/_win32: ERROR -> Data comes under a block named -> {block.name}")
                        return False

        return True

    def write_ram(self, data:bytes, addr:int=None, size:int=0) -> bool:
        """Writes data to a address. If addr is None, writes data at current addr. Size argument if present just appends data with 0"""
        if addr:
            self.current_addr = addr

        if self._CFBlock(addr, len(data) + size) == False:
            return False

        if size:
            append_ = 0
            data = data + append_.to_bytes(size, "little")

        self.ram[self.current_addr:self.current_addr + len(data)] = data
        return True

    def get_ram(self, size:int, addr:int=None) -> bytes:
        """Gets data from a part of RAM

        Args:
            size (int): The size of data to get
            addr (int, optional): The address of data (start). If None, uses current address. Defaults to None.

        Returns:
            bytes: The retrieved data
        """
        if addr:
            self.current_addr = addr

        if self._CFBlock(addr, size) == False:
            return b""

        return self.ram[self.current_addr : self.current_addr + size]

    def serialize(self) -> str:
        attribs = {"size": self.size, "proc_sector_size": self.proc_sector_size, "blocks": self.mem_blocks}
        import json
        return json.dumps(attribs)

class Process:
    """A Class for the Proc Handling (DO NOT USE!)"""
    def __init__(self, mem:Memory, exec_data:Execution_Data, ProcessData:Process_Data, addr:int=0) -> None:
        """Init func
        """
        self.lock = threading.Lock()

        self.mem:Memory = mem
        self.exec_data:Execution_Data = exec_data
        self.ProcessData:Process_Data = ProcessData
        self.pid:int = ProcessData.Pid
        self.ram:bytearray = mem.ram

        self.vram_seek = addr

        # Allocation Segments
        self.data_seek:int = addr
        self.stack_seek:int = self.data_seek + exec_data.data_size + 1
        self.heap_seek:int = self.stack_seek + exec_data.stack_size + 1
        self.text_seek:int = self.heap_seek + exec_data.heap_size + 1

        self.allocs = {} # Var address mapping

    def __setattr__(self, name:str, value:int):
        if name in ["lock", "mem", "exec_data", "ProcessData", "pid", "ram", "vram_seek", "data_seek", "text_seek", "heap_seek", "stack_seek", "allocs"]:
            super().__setattr__(name, value)
            return

        with self.lock: #Ensure Thread safety
            if isinstance(value, bytes): # Text Segment
                addr = self.text_seek
                self.text_seek += len(value)
            else:
                size = len(value.to_bytes((value.bit_length() // 8) + 1, 'little'))

                if self._is_uninitialized(name):
                    addr = self.heap_seek
                    self.heap_seek += size
                else:
                    addr = self.data_seek
                    self.data_seek += size

                if addr >= self.exec_data.vmem_size:
                    print(f"Process {self.pid}: ERROR -> Memory Overflow! Process is now stopping")
                    return

            self.mem.write_ram(value.to_bytes(size, 'little'), addr)
            self.allocs[name] = addr

    def __getattr__(self, name):
        if name in ["lock", "mem", "exec_data", "ProcessData", "pid", "ram", "vram_seek", "data_seek", "text_seek", "heap_seek", "stack_seek", "allocs"]:
            raise RuntimeError(f"Unexpected lookup for {name}!")
        if name in self.allocs:
            addr = self.allocs[name]
            return int.from_bytes(self.mem.get_ram(4, addr), 'little')
        else:
            raise AttributeError(f"{name} not found!")

    def _is_uninitialized(self, name):
        stack = inspect.stack()
        for frame in stack:
            if "self" in frame.frame.f_locals:
                if frame.frame.f_locals["self"].__class__ == self.__class__:
                    return False  # Already initialized
        return True  # Uninitialized

class Process_Manager:
    "Process Handling with Scheduling"

    def __init__(self, mem:Memory):
        self.mem:Memory = mem
        self.procs:dict[Process] = {}
        self.ready_queue = queue.Queue()
        self.lock = threading.Lock()
        self.current_addr = 65

    def add_proc(self, exec_data:Execution_Data, process_data:Process_Data) -> None:
        """Create and Add a New Process"""
        with self.lock:
            proc = Process(self.mem, exec_data, process_data, self.current_addr)
            self.procs[process_data.Pid] = proc
            self.ready_queue.put(proc)
            self.current_addr += exec_data.vmem_size + 1
            print(f"Process {process_data.Pid} has been added!")

    def run_next(self) -> None:
        """Execute Next Process in Queue"""
        proc = None
        with self.lock:
            if self.ready_queue.empty():
                print("No processes left to execute!")
                return

            proc:Process = self.ready_queue.get()
            # Call func
            try:
                out = proc.ProcessData.process(*proc.ProcessData.params)
            except Exception as e:
                print (f"Process {proc.ProcessData.Pid} has returned a error ->\n{e}\n")
                self.stop_proc(proc.ProcessData.Pid)
        self.stop_proc(proc.ProcessData.Pid)

    def stop_proc(self, pid:int):
        """Stop Running process"""
        with self.lock:
            if pid in self.procs:
                proc:Process = self.procs[pid]

                cleanup_val = 0
                current_mem = self.mem.current_addr

                # Cleanup Memory
                self.mem.write_ram(cleanup_val.to_bytes(proc.exec_data.vmem_size, 'little'), proc.vram_seek)

                self.mem.current_addr = current_mem

                del self.procs[pid]

    def start_debug_server(self, host:str="127.0.0.1", port:int=65432) -> None:
        """Start a TCP server in a background thread to serve debug commands."""
        def handle_client(conn, addr) -> None:
            with conn:
                conn.sendall(b"Connected to Process Manager debug terminal.\nType 'help' for commands.\n> ")
                while True:
                    data = conn.recv(1024)
                    if not data:
                        break

                    cmd = data.decode().strip()
                    if cmd == "exit":
                        conn.sendall(b"Goodbye!\n")
                        break
                    elif cmd == "list":
                        response = self._get_process_list_str()
                    elif cmd.startswith("info "):
                        pid_str = cmd[5:].strip()
                        response = self._get_process_info_str(pid_str)
                    elif cmd == "help":
                        response = ("Commands:\n"
                                    "  list - list all processes\n"
                                    "  info <pid> - show info on process\n"
                                    "  exit - exit this debug session\n")
                    else:
                        response = "Unknown command. Type 'help'.\n"

                    conn.sendall(response.encode() + b"\n> ")

        def server() -> None:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind((host, port))
                s.listen()
                print(f"Debuuging Server running on {host}:{port}")
                while True:
                    conn, addr = s.accept()
                    threading.Thread(target=handle_client, args=(conn, addr), daemon=True).start()
        threading.Thread(target=server, daemon=True).start()

    def _get_process_list_str(self):
        with self.lock:
            if not self.procs:
                return "No processes running."
            lines = ["Processes:"]
            for pid in self.procs.keys():
                lines.append(f"  PID: {pid}")
            return "\n".join(lines)

    def _get_process_info_str(self, pid_str):
        try:
            pid = int(pid_str, 0)
        except ValueError:
            return "Invalid PID."
        with self.lock:
            proc = self.procs.get(pid)
            if not proc:
                return f"No process with PID {pid}."
            lines = [
                f"Process {pid}:",
                f"  VRAM Addr: {proc.vram_seek}",
                f"  Data Segment Start: {proc.data_seek}",
                f"  Stack Segment Start: {proc.stack_seek}",
                f"  Heap Segment Start: {proc.heap_seek}",
                f"  Text Segment Start: {proc.text_seek}",
                f"  Allocated Vars: {list(proc.allocs.keys())}",
                f"  Allocs (var -> addr): {proc.allocs}",
            ]
            return "\n".join(lines)



# Boots up AOS Kernel
# Table
__sys_info__ = __int__(0x9)# type:ignore
__drive_info__ = __int__(0x8)# type:ignore

table = {
	"filename": "AOS_Kernel.py",
	"args": [__sys_info__, __drive_info__]
}
__int__(0x51, table) #type:ignore
#AA55


"""GUI Library using SDL.

NOTE: Still under development.
"""

import sdl2.ext
from . import *

if __name__ == '__main__':
    exit()

import os
import sys

import ctypes

from typing import *

from sdl2 import *

from . import *

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.ErrorSystem.ErrorSystem import *
from phardwareitk.Extensions import *

DWMWA_USE_IMMERSIVE_DARK_MODE = "Only Windows!"
WM_SETICON = "Only Windows!"
ICON_SMALL = "Only Windows!"
ICON_BIG = "Only Windows!"

if Windows:
    DWMWA_USE_IMMERSIVE_DARK_MODE = 20
    WM_SETICON = 0x0008
    ICON_SMALL = 0
    ICON_BIG = 1

PError:bool = False
title:Optional[str] = None
taskbarWindowedIds:list[str] = []


def SafeExitSDL(msg:Optional[Union[str, int, dict, bool, list, tuple]], DestroyFunc:object, DestroyFuncParams:tuple) -> None:
    """Safe Exit for SDL.

    Args:
        msg (Optional[Union[str, int, dict, bool, list, tuple]]): Message.
        DestroyFunc (object): The DestroyFunc in gui_sdl.py.
        DestroyFuncParams (tuple): The DestroyFunc Parameters in gui_sdl.py.
    """
    DestroyFunc(*DestroyFuncParams)
    if msg:
        msg = str(msg)
        if msg != "":
            HyperOut.printH(msg, FontEnabled=True, Font=Extensions.TextFont(font_color=Extensions.Color("red")))

    sdl2.SDL_Quit()

class WidgetPack():
    """A type for a single PACKED Widget.
    """
    def __init__(self, Widget:PWidget, MainFunc:object, MainFuncParams:list) -> None:
        """INITALIZE func.

        Args:
            Widget (PWidget): The Widget class.
            MainFunc (function): The main function of the widget.
            MainFuncParams (list): The parameters of the main function.
        """
        self.Widget = Widget
        self.MainFunc = MainFunc
        self.MainFuncParams = MainFuncParams

def InitializePError() -> None:
    """Initializes Pheonix/Print Error."""
    global PError
    PError = True

def DeinitializePError() -> None:
    """Deinitializes Pheonix/Print Error."""
    global PError
    PError = False

def initialize() -> None:
    """Initializes the SDL and PHardware GUI.

    Raises:
        PheonixException: Incase SDL does not initialize properly.
    """
    if SDL_Init(SDL_INIT_VIDEO) != 0:
        raise PheonixException("Could not initialize SDL. - ["+str(SDL_GetError())+"]")

def ExtensionsInit(video:bool=True, audio:bool=False, timer:bool=False, joystick:bool=False, controller:bool=False, haptic:bool=False, sensor:bool=False, events:bool=True) -> None:
    """Initializes the SDL2 Extensions.

    Args:
        video (bool, optional): Video. Defaults to True.
        audio (bool, optional): Audio. Defaults to False.
        timer (bool, optional): Timer. Defaults to False.
        joystick (bool, optional): Joystick. Defaults to False.
        controller (bool, optional): Controller. Defaults to False.
        haptic (bool, optional): Haptic. Defaults to False.
        sensor (bool, optional): Sensor. Defaults to False.
        events (bool, optional): Events. Defaults to True.
    """
    return sdl2.ext.init(video, audio, timer, joystick, controller, haptic, sensor, events)

def GetHwnd(Title:Optional[str]=None) -> Optional[Any]:
    """Returns the windows handle for the specific window with the specified title.

    Args:
        Title (Optional[str], optional): The title of the created window, If None uses the title of the latest window. Defaults to None.

    Returns:
        Optional[Any]: Windows Handle, used in Windows OS, OR None incase of error. HWND.
    """
    if Windows:
        import ctypes
        if not Title:
            return ctypes.windll.user32.FindWindowW(None, title)
        else:
            return ctypes.windll.user32.FindWindowW(None, Title)
    else:
        if PError:
            HyperOut.printH("Not Windows OS!", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        return None

def LoadHIcon(icon:PIcon, Title:Optional[str]=None) -> Any:
    """Loads the windows handle icon for the specific window with the specified title.

    Args:
        icon (PIcon): The icon for the HWnd. NOTE: Can use 'exe' format too, it will fetch the icon of the 'exe' file.
        Title (Optional[str], optional): The title of the created window, If None uses the title of the latest window. Defaults to None.

    Returns:
        The loaded HIcon
    """
    if not Title:
        Title = title

    if Windows:
        import ctypes

        user32 = ctypes.windll.user32

        hwnd = GetHwnd(Title)

        return user32.LoadImageW(None, icon.iconPath, ctypes.c_int(1), 0, 0, ctypes.c_int(0x0003))
    else:
        if PError:
            HyperOut.printH("Not Windows OS!", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        return None

def GetNSWindow() -> None:
    """Returns the native macOS window object. Gets the latest opened window."""
    if not Darwin:
        if PError:
            HyperOut.printH("OS is not MacOS!", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        return None

    import importlib.util
    if not importlib.util.find_spec("AppKit"):
        if PError:
            HyperOut.printH("AppKit required for such process in MacOS!", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        return None
    else:
        from AppKit import NSApp # type: ignore

        return NSApp.windows()[0]

def SetBackgroundColor(window: Any, color:Color=Color("white")) -> None:
    """Sets the background color of the specified window.

    Args:
        window (Any): The window
        color (Color, optional): The color. [phardwareitk.Extensions.Color]. Defaults to Color("white").
    """
    windowSurface = SDL_GetWindowSurface(window)
    if not windowSurface:
        if PError:
            HyperOut.printH(FromBytes(SDL_GetError()) + "!PHardware -> windowSurface == None", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        return None

    r,g,b = color.color

    sdl_color = SDL_Color(r=r, g=g, b=b, a=color.alpha)
    SDL_FillRect(windowSurface, None, SDL_MapRGBA(windowSurface.contents.format, sdl_color.r, sdl_color.g, sdl_color.b, sdl_color.a))
    SDL_UpdateWindowSurface(window)

def SetSurfaceBackgroundColor(surface:Any, color:Color=Color("white")) -> None:
    """Sets the background color of the specified surface.

    Args:
        surface (Any): The surface
        color (Color, optional): The color. [phardwareitk.Extensions.Color]. Defaults to Color("white").
    """
    if not surface:
        if PError:
            HyperOut.printH("!PHardware -> surface == None", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        return None

    r,g,b = color.color

    sdl_color = SDL_Color(r=r, g=g, b=b, a=color.alpha)
    SDL_FillRect(surface, None, SDL_MapRGBA(surface.contents.format, sdl_color.r, sdl_color.g, sdl_color.b, sdl_color.a))

def AddIcon(window:Any, icon:PIcon, taskbarIncluded:bool=True, Title: Optional[str]=None) -> None:
    """Adds an icon to the window.

    Args:
        window (Any): Window.
        icon (PIcon): Icon. [phardwareitk.Extensions.PIcon]
        taskbarIncluded (bool, optional): If True, depending on Os, the taskbar window icon will change from python to the specified icon with the icon on the window. Defaults to True.
        Title (Optional[str], optional): The title of the window, If 'None' the title of the latest window will be used. Defaults to None.
    """

    global taskbarWindowedIds

    icon.SetIconSDL2(window)

    if not Title:
        Title = title

    if taskbarIncluded:
        if Windows:
            import ctypes
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(title+"x64.phardwareitk.pheonixappapi.PheonixStudios"+str(len(taskbarWindowedIds)))

def WindowSurface(window:Any) -> Any:
    """Returns the SDL2 window surface.

    Args:
        window (Any): The window returned by 'CreateWindow' Function.

    Returns:
        Any: Window surface.
    """
    return SDL_GetWindowSurface(window)

def LoadBMP(image:str="") -> Any:
    """Loads the specified BMP image.

    Args:
        image (str, optional): The image path in string. Defaults to "".

    Returns:
        Any: The returned value of SDL2.'LoadBMP' Function.
    """
    return SDL_LoadBMP(ToBytes(image))

def UpdateWindow(window:Any) -> None:
    """Updates the window.

    Args:
        window (Any): The window.
    """
    SDL_UpdateWindowSurface(window)

def BlitSurfaceScaled(src:Any, srcrect:Any, dst:Any, dstrect:Any, scaleMode:Any) -> bool:
    """_summary_

    Args:
        src (Any): the SDL_Surface structure to be copied from.
        srcrect (Any): the SDL_Rect structure representing the rectangle to be copied, or NULL to copy the entire surface.
        dst (Any): the SDL_Surface structure that is the blit target.
        dstrect (Any): the SDL_Rect structure representing the target rectangle in the destination surface, or NULL to fill the entire destination surface.
        scaleMode (Any): the SDL_ScaleMode to be used.
    """
    return SDL_BlitScaled

def BlitSurface(src:Any, srcrect:Any, dst:Any, dstrect:Any) -> bool:
    """Same as SDL2.BlitSurface.

    Args:
        src (Any): the SDL_Surface structure to be copied from.
        srcrect (Any): the SDL_Rect structure representing the rectangle to be copied, or NULL to copy the entire surface.
        dst (Any): the SDL_Surface structure that is the blit target.
        dstrect (Any): the SDL_Rect structure representing the x and y position in the destination surface, or NULL for (0,0). The width and height are ignored, and are copied from srcrect. If you want a specific width and height, you should use BlitSurfaceScaled().
    """
    return SDL_BlitSurface(src, srcrect, dst, dstrect)

def FreeSurface(image:Any) -> None:
    """Frees the specified BMP image.

    Args:
        image (Any): The image returned from 'LoadBMP' function.
    """
    SDL_FreeSurface(image)

def CreateWindow(name:str="PHardware GUI", width:int=800, height:int=400, WindowPos:Union[None, Any]=SDL_WINDOWPOS_CENTERED, WindowType:Optional[Any]=None, x:int=0, y:int=0) -> Any:
    """Creates a window.

    Args:
        name (str, optional): The Title of the window. Defaults to "PHardware GUI".
        width (int, optional): The width of the window. Defaults to 800.
        height (int, optional): The height of the window. Defaults to 400.
        WindowPos (Union[None, Any], optional): The position of the window. Only accept SDL_WINDOWPOS_[Something]. Defaults to SDL_WINDOWPOS_CENTERED
        WindowType (Optional[Any], optional): If not NONE, The provided SDL_WINDOW_[type] will be used for the window type. Example -> SDL_WINDOW_FULLSCREEN or SDL_WINDOW_BORDERLESS. None for normal. Defaults to None
        x (int): If WindowType is not None, this will be used as the X position. Defaults to 0.
        y (int): If WindowType is not None, this will be used as the Y position. Defaults to 0.

    Raises:
        PheonixException: Incase window creation failed.

    Returns:
        Any: Window Handle
    """
    global title

    title = name

    if not WindowType:
        if WindowPos == None:
            WindowPos = SDL_WINDOWPOS_CENTERED

        name = ToBytes(name)

        window = SDL_CreateWindow(name, WindowPos, WindowPos, width, height, SDL_WINDOW_SHOWN)
        if not window:
            raise PheonixException("Could not create window. - ["+str(SDL_GetError())+"]")
        else:
            return window
    else:
        if WindowPos == None:
            WindowPos = SDL_WINDOWPOS_CENTERED

        ExtensionsInit()
        import sdl2.ext
        window = sdl2.ext.Window(name, size=(width, height), position=(x, y), flags=WindowType)
        return window

def Quit() -> None:
    """Quits the Window.
    """
    SDL_Quit()

def DestroyWindow(window:Any, icon:Optional[PIcon]=None, renderer:Optional[SDL_Renderer]=None) -> None:
    """Destroys the Window and any remaining resources.

    Args:
        window (Any): The Window.
        icon (Optional[PIcon], optional): If any, The icon used. Defaults to None.
        renderer (Optional[SDL_Renderer], optional): If any, the renderer used. Defaults to None.
    """
    if renderer:
        SDL_DestroyRenderer(renderer)
    if icon:
        icon.Free()

    return SDL_DestroyWindow(window)

def EventLoop(quit_OnMessage:bool=False, window:Any=None, WidgetPacks:Optional[list[WidgetPack]]=None, icon:Optional[PIcon]=None, renderer:Optional[SDL_Renderer]=None, SafeExitOnError:bool=True) -> Union[bool, SDL_Event]:
    """Checks if any events are in event queue and reteives the next event.

    Args:
        quit_OnMessage (bool, optional): If set to True, will run loop (Button and such widgets will not work), upon receiving Quit event (e.g. On Window Quit) the program will destroy the window. Else False. Defaults to False
        window (Any, optional): If quit_OnMessage set to true, the provide the window
        WidgetPacks (Optional[list[WidgetPack]]): If any, the program will provide events to the widget.
        ....: Parameter to DestroyWindow.

    Returns:
        Union[bool, SDL_Event]: True for Quit, otherwise the event.
    """
    try:
        event = SDL_Event()

        run = True
        while run:
            while SDL_PollEvent(ctypes.byref(event)) != 0:
                if WidgetPacks != None and WidgetPacks != []:
                    for i, v in enumerate(WidgetPacks):
                        if not v.MainFunc:
                            continue

                        if v.MainFuncParams == None or v.MainFuncParams == []:
                            v.MainFunc(event)
                        else:
                            params = tuple(v.MainFuncParams)
                            v.MainFunc(event, *params)

                if event.type == SDL_QUIT:
                    if quit_OnMessage:
                        if not icon and not renderer:
                            DestroyWindow(window)
                        elif icon and not renderer:
                            DestroyWindow(window, icon)
                        elif not icon and renderer:
                            DestroyWindow(window, renderer=renderer)
                        elif icon and renderer:
                            DestroyWindow(window, icon, renderer)

                        Quit()
                    else:
                        return True
                    run = False
                    break
    except Exception as e:
        if SafeExitOnError:
            SafeExitSDL(f"Safe Exit Acitvated! Error -> {e}")
        else:
            raise Exception(e)

def MakeRenderer(window:Any, index:int=-1, flags:Union[0, Any]=SDL_RENDERER_ACCELERATED) -> Any:
    """Returns a renderer.

    Args:
        window (Any): The Window.
        index (int, optional): The index of the rendering device to activate. -1 for default. Defaults to -1.
        flags (Any, optional): The flags for Renderer.

    Returns:
        Any: The renderer.
    """
    return SDL_CreateRenderer(window, index, flags)

def RenderText(renderer:SDL_Renderer, widget_rect:SDL_Rect, text:str, font:TextFont=TextFont(font_color=Color("black")), fontFile:Optional[str]=None) -> Union[SDL_Renderer, bytes]:
    """Renders text.

    Args:
        renderer (SDL_Renderer): The renderer to use.
        widget_rect (SDL_Rect): The rectangle inside which the text should be rendered. (RECT will be invisible)
        text (str): The text to be rendered.
        font (TextFont): The text font to be rendered.
        fontFile (Optional[str], optional): The Font file, if not present -> None. Defaults to None._
    """
    from sdl2 import sdlttf

    # Initialize SDL_ttf if not already done
    if not sdlttf.TTF_WasInit():
        sdlttf.TTF_Init()

    render_ = renderer
    r,g,b = font.color.color
    SDL_SetRenderDrawColor(render_, r,g,b,font.color.alpha)
    SDL_RenderClear(render_)

    font_ = None

    if Windows and not fontFile:
        font_ = sdlttf.TTF_OpenFont(ToBytes(f"C:\\Windows\\Fonts\\{font.font}.ttf"), 64)
    elif Linux and not fontFile:
        font_ = sdlttf.TTF_OpenFont(ToBytes(f"usr/share//fonts//{font.font}.ttf"), 64)
    elif Unix and not fontFile:
        font_ = sdlttf.TTF_OpenFont(ToBytes(f"usr/share/fonts//{font.font}.ttf"), 64)
    elif Unknown_os and not fontFile:
        if PError:
            HyperOut.printH("Unknown OS!", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        else:
            return b"Unknown OS"

        return None

    if not font_:
        if PError:
            HyperOut.printH(FromBytes(SDL_GetError()) + "!PHardware -> font == None", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        else:
            return ToBytes(FromBytes(SDL_GetError()) + " !PHardware -> font == None")

        return None

    if fontFile:
        font_ = sdlttf.TTF_OpenFont(ToBytes(fontFile), 64)

    color = SDL_Color(r, g, b, font.color.alpha)
    tsurface = sdlttf.TTF_RenderText_Solid(font_, ToBytes(text), color)

    if not tsurface:
        if PError:
            HyperOut.printH(FromBytes(SDL_GetError()) + "!PHardware -> Tsurface == None", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        else:
            return ToBytes(FromBytes(SDL_GetError()) + " !PHardware -> Tsurface == None")

        return None

    # Convert to correct surface format
    OptimTsurface = SDL_ConvertSurfaceFormat(tsurface, SDL_PIXELFORMAT_ARGB8888, 0)
    if not OptimTsurface:
        if PError:
            HyperOut.printH(FromBytes(SDL_GetError()) + "!PHardware -> OptimTsurface == None", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        else:
            return ToBytes(FromBytes(SDL_GetError()) + " !PHardware -> OptimTsurface == None")

        return None

    ttexture = SDL_CreateTextureFromSurface(render_, tsurface)

    if not ttexture:
        if PError:
            HyperOut.printH(FromBytes(SDL_GetError()) + "!PHardware -> Ttexture == None", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        else:
            return ToBytes(FromBytes(SDL_GetError()) + " !PHardware -> Ttexture == None")

        return None

    SDL_FreeSurface(tsurface)
    SDL_FreeSurface(OptimTsurface)

    SDL_RenderCopy(render_, ttexture, None, widget_rect)
    SDL_RenderPresent(render_)

    SDL_DestroyTexture(ttexture)

    return render_

def Delay(time:int) -> None:
    """Creates a delay of specified seconds, in that time, all the processes will be paused. Since, it is a window, normal time.sleep doesn't work.

    Args:
        time (int): The time to delay in seconds.
    """
    time = time * 1000
    SDL_Delay(time)
    return None

def TitleBarWindows(Title:Optional[str]=None, mode:Optional[str]=None, color:Color=Color("white")) -> None:
    """Modifies the orignal titleBar, to the specified color or mode. WINDOWS SPECIFIC

    Args:
        Title (Optional[str]): the title of the window, that needs to be modified, If 'None', uses the title of the main window. Defaults to None
        mode (Optional[str], optional): The mode to use. Available -> dark, light, None (Default). Defaults to None.
        color (Color, optional): The color of the titleBar, used incase 'mode' is set to None. Defaults to Color("black").
    """
    if not Title:
        Title = title

    if Windows:
        import ctypes

        HWnd = GetHwnd(Title)

        if not mode or mode == "":
            # Define COLORREF
            r, g, b = color.color
            colorref = (r | (g << 8) | (b << 16))

            ctypes.windll.dwmapi.DwmSetWindowAttribute(HWnd, 20, ctypes.byref(ctypes.c_int(colorref)), ctypes.sizeof(ctypes.c_int)) # Set Window Attrib
        else:
            mode = mode.lower()

            if mode == "dark":
                value = ctypes.c_int(1)
            else:
                value = ctypes.c_int(0)

            ctypes.windll.dwmapi.DwmSetWindowAttribute(HWnd, DWMWA_USE_IMMERSIVE_DARK_MODE, ctypes.byref(value), ctypes.sizeof(value)) # Set Window Attrib
    else:
        if PError:
            HyperOut.printH("Not Windows OS!", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        return None

def TitleBarMacOS(mode:Optional[str]=None) -> None:
    """Modifies the orignal titleBar, to the specified color or mode. MACOS SPECIFIC. DOESN't SUPPORT COLOR!

    Args:
        mode (Optional[str], optional): The mode to use. Available -> dark, light, None (Default). Defaults to None.
    """
    import importlib.util

    if not Darwin:
        if PError:
            HyperOut.printH("OS is not MacOS!", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        return None

    if not importlib.util.find_spec("AppKit"):
        if PError:
            HyperOut.printH("AppKit required for such process in MacOS!", FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
        return None
    else:
        from AppKit import NSApplication, NSApp, NSWindow, NSAppearance, NSAppearanceNameAqua, NSAppearanceNameDarkAqua # type: ignore
        nsWindow = GetNSWindow()

        if mode:
            appearence = NSAppearance.appearanceNamed_(NSAppearanceNameDarkAqua if mode.lower == "dark" else NSAppearanceNameAqua)
            nsWindow.setAppearance(appearence)
            nsWindow.display()
            return None

class TitleBar(PWidget):
    """A title bar class. Allows you to create custom title bars. To use this keep window creation to SDL_Window_BORDERLESS.

    NOTE: Still Under development, hence some bugs might occur.

    NOTE: The main action buttons do not work! please use with care! (Under Development)"""
    def __init__(self, window:Any, width:Optional[int]=None, height:int=50, title:str="PHardwareITK", bgcolor:Color=Color("black")) -> None:
        """Init func of TitleBar Class.

        NOTE: Still Under development, hence some bugs might occur.

        NOTE: The main action buttons do not work! please use with care! (Under Development)

        Args:
            window (Any): The window to draw the titleBar on.
            width (Optional[int], optional): The width of the title bar, None for using Window Size Width. Defaults to None.
            height (int, optional): The height of the title bar. Defaults to 50.
            title (str, optional): The title text. Defaults to "PHardwareITK".
            bgcolor (Color, optional): The background color of the title bar. Defaults to Color("black").
        """
        self.height = height
        self.width = width

        if not width:
            self.width = SDL_GetWindowSize(window)[0]

        self.window = window
        self.colors = {"bg": bgcolor}
        self.text = {"title": {"text": title, "font": TextFont(font_size=16, font_color=Color("white")), "x": "default", "y": "default", "enabled": True}}
        self.icons = {"close": PIcon(None, width - 30, 10, 20, 20), "minimize": PIcon(None, width - 90, 10, 20, 20), "maximize": PIcon(None, width - 60, 10, 20, 20)}
        self.interactivity = {"draggable": True}
        self.size = [SDL_GetWindowSize(window)[0], height]
        self.position = {"x": 0, "y": 0}
        self.iconSurfaces = {}
        self.iconTextures = {}

    def SetBackgroundColor(self, color:Color) -> None:
        """Changes the orignal background color of the title bar.

        Args:
            color (Color): The color.
        """
        self.colors["bg"] = color

    def DisableTitle(self) -> None:
        """Disables the title from the title bar"""
        self.text["title"]["enabled"] = False

    def SetTitle(self, title:str="PHardwareITK", font:TextFont=TextFont(font_size=16)) -> None:
        """Changes the orignal title of the title bar.

        Args:
            title (str, optional): The title. Defaults to "PHardwareITK".
            font (TextFont, optional): The font of the title. Defaults to TextFont(font_size=16).
        """
        self.text["title"]["text"] = title
        self.text["title"]["font"] = title

    def AddText(self, name:str, text:str, font:TextFont=TextFont(font_size=16), x:int=0, y:int=0, enabled:bool=True) -> None:
        """Adds a new text to the title bar at the specified locations.

        Args:
            name (str): The name of the object. (should be Unique and not previously used)
            text (str): The text of the object.
            font (TextFont, optional): The font of the object. Defaults to TextFont(font_size=16).
            x (int, optional): The X coordinate (in the titlebar). Defaults to 0.
            y (int, optional): The X coordinate (in the titlebar)The X coordinate (in the titlebar). Defaults to 0.
            enabled (bool, optional): Wether to enable it or not for the time being. Defaults to True.
        """
        try:
            self.text[name] = {"text": text, "font": font, "x": x, "y": y, "enabled": enabled}
        except Exception as e:
            if PError:
                HyperOut.printH("TitleBar.AddText -> " + e, FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
            return None

    def ModifyTitle(self, text:Optional[str]=None, font:Optional[TextFont]=None, x:Optional[int]=None, y:Optional[int]=None, enabled:Optional[bool]=None) -> None:
        """Modifies the Title of the title bar.

        Set parameter to 'None' to define no change.

        Args:
            text (Optional[str], optional): Text. Defaults to None.
            font (Optional[TextFont], optional): Font. Defaults to None.
            x (Optional[int], optional): X coordinate. (of the titlebar). Defaults to None.
            y (Optional[int], optional): Y coordinate. (of the titlebar). Defaults to None.
            enabled (Optional[bool], optional): Wether it is enabled or not. Defaults to None.
        """

        list_ = [text, "text", font, "font", x, "x", y, "y", enabled, "enabled"]

        skip = False
        for i, param in enumerate(list_):
            if param and not skip:
                try:
                    self.text["title"][list_[i + 1]] = param
                    skip = True
                except Exception as e:
                    if PError:
                        HyperOut.printH("TitleBar.ModifyTitle -> " + e, FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
                    return None
            if skip:
                skip = False

    def ModifyText(self, name:str, text:Optional[str]=None, font:Optional[TextFont]=None, x:Optional[int]=None, y:Optional[int]=None, enabled:Optional[bool]=None) -> None:
        """Modifies the specified Text object of the title bar.

        Set parameter to 'None' to define no change.

        Args:
            name (str): The name of the object.
            text (Optional[str], optional): Text. Defaults to None.
            font (Optional[TextFont], optional): Font. Defaults to None.
            x (Optional[int], optional): X coordinate. (of the titlebar). Defaults to None.
            y (Optional[int], optional): Y coordinate. (of the titlebar). Defaults to None.
            enabled (Optional[bool], optional): Wether it is enabled or not. Defaults to None.
        """

        list_ = [text, "text", font, "font", x, "x", y, "y", enabled, "enabled"]

        skip = False
        for i, param in enumerate(list_):
            if param and not skip:
                try:
                    self.text[name][list_[i + 1]] = param
                    skip = True
                except Exception as e:
                    if PError:
                        HyperOut.printH("TitleBar.ModifyText -> " + e, FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
                    return None
            if skip:
                skip = False

    def EnableText(self, name:str) -> None:
        """Enables the specified text object in the title bar.

        Args:
            name (str): The name of the object.
        """
        return self.ModifyText(name, enabled=True)

    def EnableTitle(self) -> None:
        """Enables the title in the title bar.
        """
        return self.ModifyTitle(enabled=True)

    def SetIcon(self, name:str, icon:PIcon) -> None:
        """Adds a new icon to the title bar at the specified x, y coordinates.

        Args:
            name (str): The name of the icon object. (Should be Unique and not previously used)
            icon (PIcon): The icon itself.
        """
        try:
            self.icons[name] = icon
        except Exception as e:
            if PError:
                HyperOut.printH("TitleBar.SetIcon -> " + e, FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
            return None

    def RemoveIcon(self, name:str) -> None:
        """Removes the specified icon from the title bar.

        Args:
            name (str): The name of the icon to remove.
        """
        try:
            del self.icons[name]
        except Exception as e:
            if PError:
                HyperOut.printH("TitleBar.RemoveIcon -> " + e, FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
            return None

    def ModifyInteractivityOptions(self, draggable:Optional[bool]=None, close_action:Optional[bool]=None, on_click:Optional[bool]=None) -> None:
        """Modifies the pre-made options for the titlebar.

        NOTE: Use 'None' to define no change.

        NOTE: Be Careful while using this, incase you set close_action to 'False', try to use some kind of program like task manager to kill the python process.

        Args:
            draggable (Optional[bool], optional): Is it draggable. Defaults to None.
            close_action (Optional[bool], optional): Can it be closed. Defaults to None.
            on_click (Optional[bool], optional): Is it clickable. Defaults to None.
        """

        list_ = [draggable, "draggable", close_action, "close_action", on_click, "on_click"]

        skip = False
        for i, v in enumerate(list_):
            if v and not skip:
                try:
                    self.interactivity[list_[i + 1]] = v
                    skip = True
                except Exception as e:
                    if PError:
                        HyperOut.printH("TitleBar.ModifyInteractivityOptions -> " + e, FontEnabled=True, Flush=True, TextFont=TextFont(font_color=Color("red")))
                    return None
            if skip:
                skip = False

    def LoadIcons(self, renderer:SDL_Renderer):
        """Loads the icons for drawing."""
        if len(self.icons) <= 0:
            return None

        for key, v in self.icons.items():
            if v.iconPath:
                self.iconSurfaces[key] = v.LoadImageSDL2()
                if self.iconSurfaces[key]:
                    self.iconTextures[key] = SDL_CreateTextureFromSurface(renderer, self.iconSurfaces[key])

    def Draw(self, renderer:SDL_Renderer) -> WidgetPack:
        """Draws the title bar based on the information.

        Args:
            renderer (SDL_Renderer): The renderer to use.

        Returns:
            WidgetPack: Pass it as a list into the EventLoop."""
        bgColor = self.colors["bg"]
        SDL_SetRenderDrawColor(renderer, *bgColor.color, bgColor.alpha)
        SDL_RenderFillRect(renderer, SDL_Rect(0, SDL_GetWindowSize(self.window)[1], self.size["width"], self.size["height"]))

        # Draw Title if any
        if "title" in self.text.keys():
            if self.text["title"]["enabled"]:
                titleSurf = None
                if self.text["title"]["x"] and self.text["title"]["y"] == "default":
                    titleSurf = RenderText(renderer, SDL_Rect(self.size["width"] // 2 - 50, SDL_GetWindowSize(self.window)[1] + (self.height // 2), len(self.text["title"]["text"]) * 2, 50), self.text["title"]["text"], self.text["title"]["font"])
                elif self.text["title"]["x"].isdigit() and self.text["title"]["y"].isdigit():
                    titleSurf = RenderText(renderer, SDL_Rect(self.text["title"]["x"], self.text["title"]["y"], len(self.text["title"]["text"]) * 2, 50), self.text["title"]["text"], self.text["title"]["font"])
                if isinstance(titleSurf, bytes):
                    return None

        for key, val in self.text.items():
            if key and val:
                if isinstance(val, dict):
                    keySurf = RenderText(renderer, SDL_Rect(self.text[key]["x"], self.text[key]["y"], len(self.text[key]["text"]) * 2, 50), self.text[key]["text"], self.text[key]["font"])
                    if isinstance(keySurf, bytes): return None

        # Draw icons
        if len(self.icons) > 0:
            for iconName, icon in self.icons.items():
                if icon:
                    if icon.iconPath:
                        SDL_RenderCopy(renderer, self.iconTextures[iconName], None, SDL_Rect(icon.x, icon.y, icon.width, icon.height))

        SDL_RenderPresent(renderer)

        return WidgetPack(TitleBar, lambda event: self._onevent(event), [])

    def _onevent(self, event:SDL_Event) -> None:
        """Do not call directly, Recommended to call via EVENT LOOP!"""

        # Interactivity stuff
        if "draggable" in self.interactivity.keys():
            self._handle_drag()

    def _handle_drag(self) -> None:
        """Private func do not use!"""
        if not hasattr(self, "_drag_state"):
            # Init drag state
            self._drag_state = {"dragging": False, "off_x": 0, "off_y": 0}

        mouse_event = self._get_mouse_event()

        if mouse_event["LEFTbutton_down"]:
            mouseX, mouseY = mouse_event["pos"]
            if 0 <= mouseX <= self.size["width"] and 0 <= mouseY <= self.size["height"]:
                self._drag_state["dragging"] = True
                self._drag_state["off_x"] = mouseX
                self._drag_state["off_y"] = mouseY

        elif mouse_event["LEFTbutton_up"]:
            # Stop Dragging
            self._drag_state["dragging"] = False

        elif self._drag_state["dragging"]:
            # Update position while dragging
            new_mouse_x, new_mouse_y = mouse_event["pos"]
            delta_x = new_mouse_x - self._drag_state["off_x"]
            delta_y = new_mouse_y - self._drag_state["off_y"]

            # Adjust title bar position
            self.position["x"] += delta_x
            self.position["y"] += delta_y

            # Update the drag offset
            self._drag_state["off_x"] = new_mouse_x
            self._drag_state["off_y"] = new_mouse_y

            # Move the window as well
            SDL_SetWindowPosition(self.window, self.position["x"], self.position["y"])

    def _get_mouse_event(self) -> dict[str, Union[bool, tuple[int, int]]]:
        """Private func do not use!"""
        mouse_state = SDL_GetMouseState()
        button_down = mouse_state & SDL_BUTTON(SDL_BUTTON_LEFT)
        button_up = not button_down
        mouseX, mouseY = SDL_GetMouseState(None, None)

        return {
            "LEFTbutton_down": bool(button_down),
            "LEFTbutton_up": bool(button_up),
            "pos": (mouseX, mouseY)
        }

class Button(PWidget):
    """A Button widget

    Args:
        PWidget (PWidget): phardwareitk.Extensions.PWidget
    """

    def __init__(self, x:int, y:int, width:int, height:int, label:str="ClickMe", bg_color:Color=Color("gray"), Font:TextFont=TextFont()) -> None:
        """Initialize the Button.

        Args:
            x (int): X coordinate.
            y (int): Y coordinate.
            width (int): Width.
            height (int): Height.
            label (str, optional): Label for the button. Defaults to "Button".
            bg_color (Color, optional): Background-Color. Defaults to Color("gray").
            Font (TextFont, optional): Font for the label. Defaults to TextFont().
        """

        super().__init__(x, y, width, height, 0, bg_color, True, Font, label)

        self.callback:object = None
        self.params:tuple = None

    def Draw(self, window:Any) -> SDL_Renderer:
        """Draws the button widget.

        Args:
            window (Any): The window to draw the widget.
        """
        surface = WindowSurface(window)
        pixel_format = surface.contents.format
        r,g,b = self.bgColor.color
        mapped_color = SDL_MapRGBA(pixel_format, r, g, b, self.bgColor.alpha)

        SDL_FillRect(surface, SDL_Rect(self.x, self.y, self.width, self.height), mapped_color)

        return RenderText(window, SDL_Rect(self.x, self.y, self.width, self.height), self.text, self.textFont) if self.text else None

    def onClick(self, callback:object, params:Optional[tuple]=None) -> WidgetPack:
        """Runs the following function if Mouse is clicked upon the button.

        Args:
            callback (object): The callback function.
            params (Optional[tuple], optional): If any, parameters for the callback. Defaults to None.

        Returns:
            WidgetPack: Use it inside the EventLoop.
        """
        self.callback = callback
        self.params = params

        return WidgetPack(Button, lambda event: self._onevent(event), [])

    def _onevent(self, event:SDL_Event) -> None:
        """Event handler. DO NOT CALL DIRECTLY. RECOMMENDED TO CALL VIA EVENT_LOOP.
        """
        if event.type == SDL_MOUSEBUTTONDOWN:
            mx = ctypes.c_int(0)
            my = ctypes.c_int(0)
            SDL_GetMouseState(ctypes.byref(mx), ctypes.byref(my))

            mx = mx.value
            my = my.value

            if self.x <= mx <= self.x + self.width and self.y <= my <= self.y + self.height:
                if self.callback:
                    if not self.params:
                        self.callback()
                    else:
                        self.callback(*self.params)

class Label(PWidget):
    """Creates a label.
    """
    def __init__(self, label:str, x:int=0, y:int=0, width:int=5, height:int=5, font:TextFont=TextFont()) -> None:
        """INIT Function for Label class.

        Args:
            label (str): The Text.
            x (int, optional): The X coordinate. Defaults to 0.
            y (int, optional): The Y coordinate. Defaults to 0.
            width (int, optional): The width of the label. Defaults to 5.
            height (int, optional): The height of the label. Defaults to 5.
            font (TextFont, optional): The font of the label. Defaults to TextFont().
        """
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.font = font
        self.label = label

    def Draw(self, window:Any) -> SDL_Renderer:
        """Draws the label on the screen.

        Args:
            window (Any): The window to draw on.

        Returns:
            SDL_Renderer: The renderer object. Should be passed onto the list of renderers in DestroyWindow function.
        """
        surface = WindowSurface(window)
        pixel_format = surface.contents.format
        r,g,b = self.font.background_color
        mapped_color = SDL_MapRGBA(pixel_format, r, g, b, self.font.background_color.alpha)

        SDL_FillRect(surface, SDL_Rect(self.x, self.y, self.width, self.height), mapped_color)

        return RenderText(window, SDL_Rect(self.x, self.y, self.width, self.height), self.label, self.font) if self.label else None

class Quadrilateral(PWidget):
    """Creates a Quadrilateral."""
    def __init__(self, width:int=5, height:int=8, x:int=0, y:int=0, color:Color=Color("white")) -> None:
        """INIT function of Quadrilateral class.

        Args:
            width (int, optional): The width of the quadrilateral. Defaults to 5.
            height (int, optional): The height of the quadrilateral. Defaults to 8.
            x (int, optional): The X coordinate. Defaults to 0.
            y (int, optional): The Y coordinate. Defaults to 0.
            color (Color, optional): The color of the quadrilateral. Defaults to Color("white").
        """
        self.width = width
        self.height = height
        self.x = x
        self.y = y
        self.color = color

    def Draw(self, window:Any) -> None:
        """Draws the quadrilateral.

        Args:
            window (Any): The window to draw the quadrilateral.
        """
        surface = WindowSurface(window)
        pixelFormat = surface.contents.format
        r,g,b = self.color.color
        mappedColor = SDL_MapRGBA(pixelFormat, r, g, b, self.color.alpha)

        SDL_FillRect(surface, SDL_Rect(self.x, self.y, self.width, self.height), mappedColor)
        return None

class Shape(PWidget):
    """Creates a shape."""
    def __init__(self, vertex:tuple, x:int=0, y:int=0, wireframe:bool=False, bgcolor:Color=Color("white"), color:Color=Color("white")) -> None:
        """INIT function of Shape Class.

        Args:
            vertex (tuple): The vertexes of the shape.
            x (int, optional): X coordinate. ADDS TO VERTEX POS, FOR ACCURATE VERTEX POS KEEP 0. Defaults to 0.
            y (int, optional): Y coordinate. ADDS TO VERTEX POS, FOR ACCURATE VERTEX POS KEEP 0. Defaults to 0.
            wireframe (bool, optional): If True, the shape will be a wireframe shape. Defaults to False.
            bgcolor (Color, optional): The background color of the shape. Defaults to Color("white").
            color (Color, optional): The color of the lines. Defaults to Color("white").
        """
        self.vertex = vertex
        self.x = x
        self.y = y
        self.wireframe = wireframe
        self.bgcolor = bgcolor
        self.color = color

    def Draw(self, renderer:SDL_Renderer) -> SDL_Renderer:
        """Draws the Shape.

        Args:
            renderer (SDL_Renderer): The renderer to use.

        Returns:
            SDL_Renderer: The renderer of the shape.
        """
        r, g, b = self.color.color

        SDL_SetRenderDrawColor(renderer, r, g, b, self.color.alpha)

        if self.wireframe:
            numPoints = len(self.vertex)

            for i in range(1, numPoints):
                x1, y1 = self.vertex[i]
                x2, y2 = self.vertex[(i + 1) % numPoints] # Wrap around to the first point
                SDL_RenderDrawLine(renderer, x1 + self.x, y1 + self.y, x2 + self.x, y2 + self.y)
        else:
            # Filled shape rendering
            r, g, b = self.bgcolor.color
            vertex_list = []

            for v in range(0, len(self.vertex) - 1):
                vx, vy = self.vertex[v]
                vertex_list.append(
                    SDL_Vertex(
                        position=sdl2.SDL_FPoint(vx + self.x, vy + self.y),
                        color=sdl2.SDL_Color(r, g, b, self.bgcolor.alpha),
                        tex_coord=sdl2.SDL_FPoint(0, 0)  # No texture coordinates for solid colors
                    )
                )

            # Convert to SDL_Vertex array
            vert_array = (SDL_Vertex * len(vertex_list))(*vertex_list)
            SDL_RenderGeometry(renderer, None, vert_array, len(vertex_list), None, 0)

        # Present the renderer (required to display the drawing)
        SDL_RenderPresent(renderer)
        return renderer

class AdvQuadrilateral(PWidget):
    """Creates a Advanced Quadrilateral."""
    def __init__(self, p1:int, p1w:int, p2:int, p2w:int, p3:int, p3w:int, p4:int, p4w:int, x:int=0, y:int=0, wireframe:bool=False, bgcolor:Color=Color("white"), color:Color=Color("white")) -> None:
        """INIT function of AdvQuadrilateral Class.

        Args:
            p1 (int): Vertex X.
            p1w (int): Vertex Y.
            ....... Continued
            x (int, optional): X coordinate. ADDS TO VERTEX POS, FOR ACCURATE VERTEX POS KEEP 0. Defaults to 0.
            y (int, optional): Y coordinate. ADDS TO VERTEX POS, FOR ACCURATE VERTEX POS KEEP 0. Defaults to 0.
            wireframe (bool, optional): Whether to make the shape wireframe. Defaults to False.
            bgcolor (Color, optional): The background color of the shape. Defaults to Color("white").
            color (Color, optional): The color. Defaults to Color("white").
        """
        self.p1 = p1
        self.p1w = p1w
        self.p2 = p2
        self.p2w = p2w
        self.p3 = p3
        self.p3w = p3w
        self.p4 = p4
        self.p4w = p4w
        self.x = x
        self.y =y
        self.wireframe = wireframe
        self.bgcolor = bgcolor
        self.color = color

    def Draw(self, renderer:SDL_Renderer) -> SDL_Renderer:
        """Draws the shape

        Args:
            renderer (SDL_Renderer): The renderer to draw using.

        Returns:
            SDL_Renderer: The renderer of the shape.
        """
        return Shape(((self.p1, self.p1w), (self.p2, self.p2w), (self.p3, self.p3w), (self.p4, self.p4w)), self.x, self.y, self.wireframe, self.bgcolor, self.color).Draw(renderer)




import os
import sys

module_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")

if not module_path in sys.path:
    sys.path.append(module_path)

from phardwareitk.PENV.shared import *

from phardwareitk.PENV.Drivers.display import *

# Never Give UP!!!!

class Framebuffer:
    """Bro i lost my mind building this"""

    def __init__(self, width: int, height: int, depth: int = 3, classNameWin32:str="Static", winNameWin32:str="Framebuffer") -> None:
        """Initialize the framebuffer with specified dimensions and depth."""
        self.width: int = width
        self.height: int = height
        self.depth: int = depth
        self.driver = None
        self.win32, self.posix, self.unknown, self.host_os = get_os()
        self.running = True  # Control flag for the message loop thread
        if self.posix:
            self.driver: LinuxFrameBufferDriver = LinuxFrameBufferDriver(width, height, depth)
        elif self.win32:
            self.driver: WindowsFrameBufferDriver = WindowsFrameBufferDriver(width, height, depth, classNameWin32, winNameWin32)

    def write_pixel(self, x=0, y=0, r=0, g=0, b=0) -> None:
        """Write a pixel to the framebuffer at the specified coordinates with the given RGB color."""
        return self.driver.wp(x, y, r, g, b)

    def clear(self, r=0, g=0, b=0) -> None:
        """Clear the framebuffer with the specified RGB color."""
        return self.driver.cl(r, g, b)

    def flush(self) -> None:
        """Flush the framebuffer to the display."""
        return self.driver.flush()

    def handle_events(self) -> None:
        """Handle events for the framebuffer driver."""
        if self.driver is not None:
            self.driver.msg_loop()

    def lserialize(self) -> list:
        """Serialize the framebuffer dimensions into a list."""
        return [self.width, self.height, self.depth]

    def delete(self) -> None:
        """Delete the framebuffer and clean up resources."""
        self.running = False  # Signal
        if self.driver is not None:
            del self.driver




import sys
import os

baseDir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")

if not baseDir in sys.path:
    sys.path.append(baseDir)

from GPU import *
set_gpu_type()
from Extensions import *
from CLI.cliToolKit import *

import time

clinfo = True

def stress_test_torch(duration=60, tensor_size=3000):
    """Stress test for CUDA or ROCm GPUs using PyTorch."""
    import torch

    if not torch.cuda.is_available():
        Text.WriteText("No CUDA/ROCm GPU detected. Exiting PyTorch stress test.", FontEnabled=True, Font=TextFont(font_color=Color("red")))
        exit(1)

    device = torch.device("cuda")
    start_time = time.time()
    elapsed_time = 0
    iterations = 0
    ranking = 0
    pps = 0
    sl = ""

    while elapsed_time < duration:
        try:
            # GPU-intensive operations
            A = torch.randn((tensor_size, tensor_size), device=device)
            B = torch.randn((tensor_size, tensor_size), device=device)
            C = torch.matmul(A, B)
            D = torch.sin(C) + torch.cos(C)
            del A, B, C, D
            torch.cuda.empty_cache()

            elapsed_time = time.time() - start_time
            iterations += 1
        except RuntimeError as e:
            Text.WriteText(e, FontEnabled=True, Font=TextFont(font_color=Color("red")))
            torch.cuda.empty_cache()
            break

    memory_allocated = torch.cuda.memory_allocated()  # Memory usage in bytes
    total_memory = torch.cuda.get_device_properties(0).total_memory
    memory_percentage = (memory_allocated / total_memory) * 100

    # Simple heuristic for stress level
    if memory_percentage > 90:
        sl = "Very High"
    elif memory_percentage > 70:
        sl = "High"
    elif memory_percentage > 50:
        sl = "Medium"
    else:
        sl = "Low"

    ranking = iterations / elapsed_time

    pps = tensor_size ** 2 / elapsed_time

    return iterations, elapsed_time, ranking, sl, pps, memory_allocated, memory_percentage, total_memory

def stress_test_opencl(duration=60, tensor_size=1024):
    """Stress test for OpenCL-supported GPUs."""
    import pyopencl as cl

    if not cl:
        Text.WriteText("PyOpenCL is not installed. Exiting OpenCL stress test.", FontEnabled=True, Font=TextFont(font_color=Color("red")))
        exit(1)

    platforms = cl.get_platforms()
    if not platforms:
        Text.WriteText("No OpenCL platforms found. Exiting OpenCL stress test.", FontEnabled=True, Font=TextFont(font_color=Color("red")))
        exit(1)

    device = platforms[0].get_devices(device_type=cl.device_type.GPU)[0]
    context = cl.Context(devices=[device])
    queue = cl.CommandQueue(context)

    start_time = time.time()
    elapsed_time = 0
    iterations = 0
    ranking = 0
    sl = ""
    pps = 0

    # Simple kernel code
    kernel_code = """
    __kernel void matmul(__global float* A, __global float* B, __global float* C, int N) {
        int row = get_global_id(0);
        int col = get_global_id(1);
        float sum = 0;
        for (int k = 0; k < N; k++) {
            sum += A[row * N + k] * B[k * N + col];
        }
        C[row * N + col] = sum;
    }
    """
    program = cl.Program(context, kernel_code).build()
    matmul_kernel = program.matmul

    while elapsed_time < duration:
        try:
            # Allocate buffers for the matrices
            A = cl.Buffer(context, cl.mem_flags.READ_WRITE, size=tensor_size * tensor_size * 4)
            B = cl.Buffer(context, cl.mem_flags.READ_WRITE, size=tensor_size * tensor_size * 4)
            C = cl.Buffer(context, cl.mem_flags.READ_WRITE, size=tensor_size * tensor_size * 4)

            import ctypes
            # Explicitly cast tensor_size to int or size_t as expected by OpenCL
            matmul_kernel.set_args(A, B, C, ctypes.c_size_t(tensor_size))

            # Enqueue kernel and execute
            global_size = (tensor_size, tensor_size)
            cl.enqueue_nd_range_kernel(queue, matmul_kernel, global_size, None)

            # Synchronize and clear buffers
            queue.finish()
            elapsed_time = time.time() - start_time
            iterations += 1
        except cl.RuntimeError as e:
            Text.WriteText(f"OpenCL runtime error: {e}", FontEnabled=True, Font=TextFont(font_color=Color("red")))
            exit(1)

    memory_allocated = None
    memory_percentage = None

    if clinfo:
        result = subprocess.run(["clinfo"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)  # Memory usage in bytes
        out = None

        if result != 0:
            Text.WriteText(f"\n{result.stderr}", FontEnabled=True, Font=TextFont(font_color=Color("red")))
        else:
            out = result.stdout

        if out:
            for line in out.splitlines():
                if "used memory" in line.lower():
                    memory_allocated = int(line.lower().split(":")[1].strip()[0])
    else:
        memory_allocated = device.global_mem_cache_size

    total_memory = device.global_mem_size # Not accurate

    if not memory_allocated == None:
        memory_percentage = (memory_allocated / total_memory) * 100

        # Simple heuristic for stress level
        if memory_percentage > 90:
            sl = "Very High"
        elif memory_percentage > 70:
            sl = "High"
        elif memory_percentage > 50:
            sl = "Medium"
        else:
            sl = "Low"
    else:
        memory_allocated = "Unknown"
        memory_percentage = "Unknown"
        sl = "Unknown"

    ranking = iterations / elapsed_time

    pps = tensor_size ** 2 / elapsed_time

    return iterations, elapsed_time, ranking, sl, pps, memory_allocated, memory_percentage, total_memory


Screen.ClearScreen()
Cursor.MoveCursor(0, 0)
Text.WriteText(f"Initializing Stress-Test for GPU:[{get_gpu_type()}] ...", FontEnabled=True, Font=TextFont(font_color=Color("yellow")))
Cursor.MoveCursor(0, 3)
Text.WriteText("Checking Required Modules..." , FontEnabled=True, Font=TextFont(font_color=Color("yellow")))

from importlib import util

time.sleep(1)

rq_mod = ""

if get_gpu_type().lower() == "nvidia":
    rq_mod = "torch"
elif get_gpu_type().lower() in ["intel uhd", "amd"]:
    rq_mod = "pyopencl"

if util.find_spec("torch") and get_gpu_type().lower() == "nvidia":
    Cursor.MoveCursor(0, 3)
    Text.WriteText("All Required Modules Present!" , FontEnabled=True, Font=TextFont(font_color=Color("green")))
elif util.find_spec("pyopencl") and get_gpu_type().lower() in ["intel uhd", "amd"]:
    Cursor.MoveCursor(0, 3)
    Text.WriteText("All Required Modules Present!", endl="\n", FontEnabled=True, Font=TextFont(font_color=Color("green")))
    try:
        out = subprocess.run(["clinfo"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
        if out != 0:
            Text.WriteText(f"[clinfo] not present. Download [clinfo] to run this program with better accuracy on [{get_gpu_type()}]", FontEnabled=True, Font=TextFont(font_color=Color("yellow")))
            clinfo = False
            time.sleep(4)
    except FileNotFoundError:
        Text.WriteText(f"[clinfo] not present. Download [clinfo] to run this program with better accuracy on [{get_gpu_type()}]", FontEnabled=True, Font=TextFont(font_color=Color("yellow")))
        clinfo = False
        time.sleep(4)
else:
    Cursor.MoveCursor(0, 3)
    Text.WriteText(f"Required Module [{rq_mod}] not Present!" , FontEnabled=True, Font=TextFont(font_color=Color("red")))
    time.sleep(0.5)
    Cursor.MoveCursor(0, 4)
    out = Text.InputText(f"Do you want to attempt auto install [{rq_mod}]? (y/n): " , endl=None, FontEnabled=True, Font=TextFont(font_color=Color("blue")))
    if out.lower() == "n":
        Screen.ClearScreen()
        Cursor.MoveCursor(0, 0)
        exit(1)
    elif out.lower() == "y":
        try:
            os.system(f"pip install {rq_mod}")
        except Exception as e:
            Cursor.MoveCursor(0, 4)
            Text.WriteText("Exception during auto install ->\n", e, FontEnabled=True, Font=TextFont(font_color=Color("red")))
            exit(1)

        Screen.ClearScreen()
        Cursor.MoveCursor(0, 0)
        Text.WriteText(f"Initialized Stress-Test for GPU:[{get_gpu_type()}]: Completed", FontEnabled=True, Font=TextFont(font_color=Color("green")))
        Cursor.MoveCursor(0, 2)
        Text.WriteText("Re-Running Program...", FontEnabled=True, Font=TextFont(font_color=Color("yellow")))
        time.sleep(1)
        try:
            os.system(f"python {os.path.relpath(__file__)}")
        except Exception:
            os.system(f"py {os.path.relpath(__file__)}")
    else:
        Screen.ClearScreen()
        Cursor.MoveCursor(0, 0)
        exit(1)

Screen.ClearScreen()
Cursor.MoveCursor(0, 0)
Text.WriteText(f"Initialized Stress-Test for GPU:[{get_gpu_type()}] ...", FontEnabled=True, Font=TextFont(font_color=Color("green")))

Cursor.MoveCursor(0, 2)
Text.WriteText("Starting Stress-Test...", FontEnabled=True, Font=TextFont(font_color=Color("yellow")))
Cursor.MoveCursor(0, 3)

iterations = 0
elapsed_time = 0
ranking = 0
stress_levels = 0
pps = 0
memAll = None
memPer = None
TotalMem = None
type_ = ""

dur = Text.InputText("Duration of test in seconds (Leave for default 60 seconds): ", endl=None, FontEnabled=True, Font=TextFont(font_color=Color("blue")))
if dur == "":
    dur = "60"
while not dur.isdigit():
    Screen.ClearLine(3)
    Cursor.MoveCursor(0, 3)
    dur = Text.InputText("Invalid input! Please enter a number or leave for default: ", endl=None, FontEnabled=True, Font=TextFont(font_color=Color("blue")))
    if dur == "":
        dur = "60"

dur = int(dur)

Screen.ClearLine(3)
Cursor.MoveCursor(0, 3)
Text.WriteText(f"Duration set to [{dur}] seconds.", endl="\n", FontEnabled=True, Font=TextFont(font_color=Color("green")))

tensor = Text.InputText("Tensor of test (Leave for default 1024/3000): ", endl=None, FontEnabled=True, Font=TextFont(font_color=Color("blue")))
if tensor == "":
    if get_gpu_type().lower() == "nvidia":
        tensor = "3000"
    elif get_gpu_type().lower() in ["intel uhd", "amd"]:
        tensor = "1024"
while not tensor.isdigit():
    Screen.ClearLine(4)
    Cursor.MoveCursor(0, 4)
    tensor = Text.InputText("Invalid input! Please enter a number or leave for default: ", endl=None, FontEnabled=True, Font=TextFont(font_color=Color("blue")))
    if tensor == "":
        if get_gpu_type().lower() == "nvidia":
            tensor = "3000"
        elif get_gpu_type().lower() in ["intel uhd", "amd"]:
            tensor = "1024"

tensor = int(tensor)

Screen.ClearLine(4)
Cursor.MoveCursor(0, 4)
Text.WriteText(f"Tensor set to [{tensor}].", endl="\n", FontEnabled=True, Font=TextFont(font_color=Color("green")))

try:
    if get_gpu_type().lower() == "nvidia":
        tensor = 3000
        iterations, elapsed_time, ranking, stress_levels, pps, memAll, memPer, TotalMem = stress_test_torch(dur, tensor)
        type_ = "PyTorch"
    elif get_gpu_type().lower() in ["intel uhd", "amd"]:
        iterations, elapsed_time, ranking, stress_levels, pps, memAll, memPer, TotalMem = stress_test_opencl(dur, tensor)
        type_ = "PyOpenCL"
except Exception as e:
    Text.WriteText(e, FontEnabled=True, Font=TextFont(font_color=Color("red")))
    exit(1)

time.sleep(1)

Screen.ClearScreen()
Cursor.MoveCursor(0, 0)
Text.WriteText(f"Initialized Stress-Test for GPU:[{get_gpu_type()}]: Completed", FontEnabled=True, Font=TextFont(font_color=Color("green")))

Cursor.MoveCursor(0, 2)
Text.WriteText(f"Stress-Test: Completed using -> [{type_}]\nResults ->\n\tIterations: {iterations}\n\tElapsed Time: {elapsed_time:.2f} Seconds\n\tStress Levels: {stress_levels}\n\tPixel Per Second (Estimated): {pps}\n\tTotal Memory: {TotalMem / (1024**3):.2f} GB\n\tMemory Usage: {memPer:.2f}%\n\tMemory Allocated: {memAll / (1024**3):.2f} GB\n\n\tOverall Ranking: {ranking}", FontEnabled=True, Font=TextFont(font_color=Color("green")))


"""Memory Allocation and such functions for python.
"""

import platform
import os
import sys

module_path:str = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")

if not module_path in sys.path:
    sys.path.append(module_path)

os_:str = platform.system()
os_ = os_.lower()
win32:bool = False
posix:bool = False
unknown:bool = True

posix_os = [
    "linux", "unix", "euleros", "macos", "openbsd", "plan9", "vxworks", "minix",
    "ubuntu", "debian", "fedora", "arch", "aix", "hp-ux", "solaris",
    "netbsd", "freebsd", "dragonflybsd", "qnx", "illumos", "truenas",
    "redhat", "centos", "rocky linux", "almalinux", "suse", "gentoo",
    "slackware", "mandriva", "scientific linux", "oracle linux"
]

if os_ == "windows":
    win32 = True
    unknown = False
    posix = False
elif os_ in posix_os:
    posix = True
    unknown = False
    win32 = False
else:
    unknown = True
    win32 = False
    posix = False

def force_os(_os:str, posix_based_os:bool=False) -> None:
    """Forces a OS so that the script will follow that specific os

    Args:
        _os (str): The OS you want to force.
    """
    _os = _os.lower()

    if _os == "windows":
        win32 = True
        posix = False
        unknown = False
    elif _os in posix_os:
        posix = True
        unknown = False
        win32 = False
    else:
        os_ = _os
        unknown = False
        win32 = False
        posix = False
        if posix_based_os:
            posix = True

if win32:
    #from phardwareitk.Memory._platform import _win32 as memory
    pass

if posix:
    #from phardwareitk.Memory._platform import _posix as memory
    pass

if unknown:
    #raise Exception("Cannot get OS try to force OS instead!")
    pass

if os_ != "" and (not unknown and (not win32 and not posix)):
    print("Sorry this is a unsupported or unimplemented OS, Exiting...")
    exit(1)



"""GUI Library, from PHardwareITK itself"""

import platform
import os
import sys
import ctypes

from typing import *

from phardwareitk.GPU._base import BaseGPUD
from phardwareitk.GUI.PheonixIon.types import *

class NoCompatibleGraphicsAPIFound(Exception):
    def __init__(self, message="No compatible graphics API found"):
        super().__init__(message)
class UnsupportedPlatform(Exception):
    def __init__(self, message="Unsupported platform for window creation"):
        super().__init__(message)

SYSTEM = platform.system().lower()

def try_load_library(lib_names:list):
    """Try to load a list of library names; return first that succeeds."""
    for name in lib_names:
        try:
            return ctypes.CDLL(name)
        except OSError:
            continue
    return None

class PIonWindow:
    """A Pheonix Ion Window Instance"""
    def __init__(self, title:str, width:int, height:int, flags=None, handle=None) -> None:
        self.title = title
        self.width = width
        self.height = height
        self.handle = handle

        self.backends = PIenum()
        self.backends.create_enum(["Win32", "X11", "Cocoa"], 1)
        self.backend = self._select_backend()

        self.flags = flags

    def _select_backend(self, backend:Union[str, PIenum]=None) -> str:
        if not backend:
            if SYSTEM == "windows":
                return self.backends.access("Win32")
            elif SYSTEM == "linux":
                return self.backends.access("X11")
            elif SYSTEM == "darwin":
                return self.backends.access("Cocoa")
            else:
                print("Unknown platform for automatic backend selection!, please specify backend manually.")
                return None
        else:
            if isinstance(backend, str):
                bkend = self.backends.access(backend)
                if bkend == None:
                    raise UnsupportedPlatform(f"Unsupported backend: {backend}")
                return bkend
            elif isinstance(backend, PIenum):
                return backend.value
            else:
                raise ValueError("Backend must be a string or PIenum instance")

class GPU_API:
    """Pheonix Ion GPU API, use the driver attribute to access functions like -
    ```python
        mygpu = GPU_API() # This is a simplification, the PheonixIon Class will return the GPU_API initialized.
        # Initialize
        mygpu.driver.init(mydisplay, mywin)
        # Clear screen
        mygpu.driver.clear(255, 0, 0, 255) # rgba
        # Destroy
        mygpu.driver.shutdown()
    ```
    """
    def __init__(self, api:str=None, driver:Any=None) -> None:
        self.api = api if api is not None else self._detect_best_api()
        self.available_api = self._get_available_apis()
        self.context_handle = None
        from phardwareitk.GPU._base import BaseGPUD
        self.driver: BaseGPUD = None

        if driver is None:
            for api in self.available_api:
                if api.lower() == "opengl":
                    from phardwareitk.GPU._gl_driver import GLDriver as _driver
                    self.driver = _driver()
                    if not api == self.api:
                        self.api = api
                    break

            else:
                raise NoCompatibleGraphicsAPIFound(f"no {self.available_api} graphics drivers found in phardwareitk, try implementing your own?")
        else:
            self.driver = driver

    def _detect_best_api(self) -> str:
        if self._is_vulkan_available():
            return "Vulkan"
        elif self._opengl_available():
            return "OpenGL"
        elif self._is_directx_available():
            return "DirectX"
        elif self._is_metal_available():
            return "Metal"
        elif self._is_opengl_es_available():
            return "OpenGL ES"
        else:
            raise NoCompatibleGraphicsAPIFound("No compatible graphics API found on this system.")

    def _get_available_apis(self) -> str:
        avail = []
        if self._is_vulkan_available():
            avail.append("Vulkan")
        if self._opengl_available():
            avail.append("OpenGL")
        if self._is_directx_available():
            avail.append("DirectX")
        if self._is_metal_available():
            avail.append("Metal")
        if self._is_opengl_es_available():
            avail.append("OpenGL ES")
        return avail

    def _is_vulkan_available(self) -> bool:
        if SYSTEM == "windows":
            return try_load_library(["vulkan-1.dll"]) is not None
        elif SYSTEM == "linux":
            return try_load_library(["libvulkan.so.1", "libvulkan.so"]) is not None
        elif SYSTEM == "darwin":
            return False  # Vulkan requires MoltenVK on macOS; assume false for now
        return False

    def _opengl_available(self) -> bool:
        if SYSTEM == "windows":
            return try_load_library(["opengl32.dll"]) is not None
        elif SYSTEM == "linux":
            return try_load_library(["libGL.so.1", "libGL.so"]) is not None
        elif SYSTEM == "darwin":
            return try_load_library(["/System/Library/Frameworks/OpenGL.framework/OpenGL"]) is not None
        return False

    def _is_directx_available(self) -> bool:
        if SYSTEM == "windows":
            return try_load_library(["d3d11.dll", "d3d12.dll"]) is not None
        return False

    def _is_metal_available(self) -> bool:
        # Placeholder for actual Metal detection logic
        return SYSTEM == "darwin"

    def _is_opengl_es_available(self) -> bool:
        if SYSTEM == "windows":
            return try_load_library(["libGLESv2.dll"]) is not None
        elif SYSTEM == "linux":
            return try_load_library(["libGLESv2.so"]) is not None
        elif SYSTEM == "darwin":
            return try_load_library(["/System/Library/Frameworks/OpenGLES.framework/OpenGLES"]) is not None
        return False

class PIonContext:
    def __init__(self, api_name: str, native_handle: object, lib_path: str, metadata: dict = None):
        self.api_name = api_name # e.g., "OpenGL", "Vulkan", "DirectX", "Metal"
        self.native_handle = native_handle # platform-specific (X11, HWND, Android Surface)
        self.lib_path = lib_path # path to the lib
        self.metadata = metadata or {} # optional: version, extensions, etc.

class PheonixIon:
    """The Main Pheonix Ion Multi-Window System, using one instance a user can create 1, 2, 3, ..., 50+ windows if they wanted."""
    def __init__(self):
        self.system:str = SYSTEM
        self.windows:list[PIonWindow] = []

        self.pionbackend = None
        self.backend = ""

        self._load_backend()

    def _load_backend(self, backend__:str=None):
        if backend__:
            backend_ = backend__.lower()
            if backend_ == "win32":
                from phardwareitk.GUI.PheonixIon import win32 as backend
                self.pionbackend = backend
                self.backend = "Win32"
            elif backend_ == "x11":
                from phardwareitk.GUI.PheonixIon import x11 as backend
                self.pionbackend = backend
                self.backend = "X11"
            elif backend_ == "cacoa":
                from phardwareitk.GUI.PheonixIon import cacao as backend
                self.pionbackend = backend
                self.backend = "Cacoa"
            else:
                raise UnsupportedPlatform(f"Unknown backend: {backend__}")

            return None

        if self.system == "windows":
            from phardwareitk.GUI.PheonixIon import win32 as backend
            self.pionbackend = backend
            self.backend = "Win32"
        elif self.system == "linux":
            from phardwareitk.GUI.PheonixIon import x11 as backend
            self.pionbackend = backend
            self.backend = "X11"
        elif self.system == "darwin":
            from phardwareitk.GUI.PheonixIon import cocoa as backend
            self.pionbackend = backend
            self.backend = "Cacoa"
        else:
            raise UnsupportedPlatform(f"Unsupported platform: {self.system}")
        return None

    def create_window(self, title:str="Pheonix Ion", width:int=800, height:int=600, flags=None) -> PIonWindow:
        """Create a window with specified parameters.
        Parameters:
            title (str): Title of the window.
            width (int): Width of the window.
            height (int): Height of the window.
            backend (str or PIenum): Backend to use for window creation. If None, auto-detects based on platform.
            flags: Additional flags for window creation (backend-specific), For example PIWin32Flags for Win32 backend.
        Returns:
            PIonWindow: The created window object.
        """
        handle = self.pionbackend.create_window(title, width, height, flags)

        win = PIonWindow(title, width, height, flags, handle)
        win._select_backend(self.backend)

        self.windows.append(win)

        return win

    def poll_events(self, win:Union[int, PIonWindow]) -> list:
        """Poll events for the created window.
        
        Parameters:
            win (int | PIonWindow): The window index or the PIonWindow class instance
            
        Returns:
            list: A List of PIonEvent from PheonixIon.types"""
        handle = self.get_native_handle(win)

        if not handle:
            return []

        return self.pionbackend.poll_events(handle)

    def destroy_window(self, win: Union[int, PIonWindow]):
        """
        Destroy a window and release its resources.

        Parameters:
            win (int | PIonWindow): Window index or PIonWindow instance.
        """
        window = self.get_native_handle(win)
        if not window:
            return
        self.pionbackend.destroy_window(window)
        self.windows = [w for w in self.windows if w.handle != window]

    def hide_window(self, win: Union[int, PIonWindow]):
        """
        Hide a window (make it invisible but not destroyed).

        Parameters:
            win (int | PIonWindow): Window index or PIonWindow instance.
        """
        window = self.get_native_handle(win)
        if window:
            self.pionbackend.hide_window(window)

    def show_window(self, win: Union[int, PIonWindow]):
        """
        Show a previously hidden window.

        Parameters:
            win (int | PIonWindow): Window index or PIonWindow instance.
        """
        window = self.get_native_handle(win)
        if window:
            self.pionbackend.show_window(window)

    def set_window_title(self, win: Union[int, PIonWindow], title: str):
        """
        Change the title of a window.

        Parameters:
            win (int | PIonWindow): Window index or PIonWindow instance.
            title (str): New window title.
        """
        window = self.get_native_handle(win)
        if window:
            self.pionbackend.set_window_title(window, title)
            w = self.get_window(win)
            if w:
                w.title = title

    def is_window_alive(self, win: Union[int, PIonWindow]) -> bool:
        """
        Check if a window is still alive (not closed/destroyed).

        Parameters:
            win (int | PIonWindow): Window index or PIonWindow instance.

        Returns:
            bool: True if the window is alive, False otherwise.
        """
        window = self.get_native_handle(win)
        if window:
            return self.pionbackend.is_window_alive(window)
        return False

    def get_window(self, win:Union[int, PIonWindow], throw_err=False) -> Optional[PIonWindow]:
        """Gets the PIonWindow based of its index, or class"""
        if isinstance(win, int):
            return self.windows[win]
        elif isinstance(win, PIonWindow):
            return win
        else:
            if throw_err: raise ValueError("Unknown type provided for parameter 'win', supported types are - int or PIonWindow")
            return None

    def get_native_handle(self, win:Union[int, PIonWindow]) -> Any:
        """Return OS-specific window handle (HWND on Win, X11 tuple, NSWindow on Mac)."""
        handle = self.get_window(win)
        if not handle: return None

        return handle.handle
    
    def get_gpu(self, win:Union[int, PIonWindow], api:Optional[str]=None, driver:Optional[BaseGPUD]=None) -> Optional[GPU_API]:
        """Returns the GPU context in form of GPU_API class"""
        window = self.get_native_handle(win)
        if window:
            return self.pionbackend.get_gpu(window, api, driver)
        return None
    
    def attach_gpu(self, win:Union[int, PIonWindow], gpu:GPU_API) -> None:
        window = self.get_native_handle(win)
        if window:
            return self.pionbackend.attach_gpu(window, gpu)
        return None
    